/*
 * @(#) $Id: particle-deposition.c 335 2006-01-02 23:20:29Z yeti $
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2005 Yeti (David Necas) <yeti@physics.muni.cz>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
 * This plug-in simulates particle deposition on a periodic surface and
 * visualizes the resulting height field.
 *
 * Please send patches, suggestions, and other invectives to me:
 * <yeti@physics.muni.cz>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
/* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "libgimp/stdplugins-intl.h"

#define PROCEDURE_NAME   "particle_deposition"
#define DATA_KEY_VALS    "plug_in_particle_deposition"

enum {
    BITS_PER_SAMPLE   = 8,
    CHANNEL_MAX_VALUE = (1U << BITS_PER_SAMPLE) - 1,

    PREVIEW_SIZE = 128,
    SCALE_WIDTH = 201,

    /* FIXME: Is it still impossible to store true unlimited strings as
     * parameters? */
    GRADIENT_NAME_SIZE = 256,
    GRADIENT_SAMPLE_SIZE = 8192,

    /* How often progressbar is updated */
    PROGRESS_STEP = 0x40000
};

enum {
    RESPONSE_REVERT = 1,
    RESPONSE_RESET
};

enum {
    RECOMPUTE_ALL   = 1 << 0,
    RECOMPUTE_COLOR = 1 << 1
};

typedef enum {
    COLORING_METHOD_GRAY     = 0,
    COLORING_METHOD_GRADIENT = 1
} ColoringMethod;

typedef struct {
    gdouble        density;
    gdouble        size_mean;
    gdouble        size_range;
    gdouble        vert_mean;
    gdouble        vert_range;
    gdouble        dir_strength;
    gdouble        dir_angle;
    gboolean       low_born;
    gboolean       sharper;
    gint           seed;
    ColoringMethod cm;
    gchar          grad[GRADIENT_NAME_SIZE];
    gboolean       grad_invert;
    gboolean       equalize_map;
    gboolean       random_seed;
} PluginValues;

typedef struct {
    GtkWidget *seed;
    GtkWidget *cm_gray;
    GtkWidget *cm_grad;
    GtkWidget *grad;
    GtkWidget *grad_invert;
    GtkWidget *equalize_map;
    GtkWidget *low_born;
    GtkWidget *sharper;
    GtkObject *density;
    GtkObject *size_mean;
    GtkObject *size_range;
    GtkObject *vert_mean;
    GtkObject *vert_range;
    GtkObject *dir_strength;
    GtkObject *dir_angle;
    GtkWidget *update;
} PluginControls;

static void     query                  (void);
static void     run                    (const gchar *name,
                                        gint nparams,
                                        const GimpParam *param,
                                        gint *nreturn_vals,
                                        GimpParam **return_vals);
static gboolean plugin_dialog          (void);
static void     coloring_method_changed(GtkWidget *widget,
                                        gpointer data);
static void     gradient_changed       (const gchar *name,
                                        gint width,
                                        const gdouble *grad_data,
                                        gboolean dialog_closing,
                                        gpointer user_data);
static void     refresh_controls       (const PluginValues *new_pvals);
static void     execute_all            (GimpDrawable *drawable);
static void     execute_color          (GimpDrawable *drawable);
static void     execute                (GimpDrawable *drawable,
                                        guint recompute);
static void     commit                 (GimpDrawable *drawable,
                                        gint chmax);
static void     commit_one_tile        (GimpPixelRgn *rgn,
                                        gint chmax,
                                        gint gradwidth);
static void     render_gradient        (void);

static GimpPlugInInfo PLUG_IN_INFO = {
    NULL,  /* init_proc  */
    NULL,  /* quit_proc  */
    query, /* query_proc */
    run,   /* run_proc   */
};

static const PluginValues pvals_defaults = {
    120.0,                    /* density      */
    0.4,                      /* size_mean    */
    0.3,                      /* size_range   */
    0.2,                      /* vert_mean    */
    0.1,                      /* vert_range   */
    0.0,                      /* dir_strength */
    0.0,                      /* dir_angle    */
    FALSE,                    /* low_born     */
    FALSE,                    /* sharper      */
    42,                       /* seed         */
    COLORING_METHOD_GRADIENT, /* cm           */
    "",                       /* grad         */
    FALSE,                    /* grad_invert  */
    FALSE,                    /* equalize_map */
    0,                        /* random_seed  */
};

static PluginControls pctrl;
static PluginValues pvals; /* = pvals_defaults; */
static PluginValues pvals_old; /* for revert */

static GtkWidget *preview_image = NULL;
static GdkPixbuf *preview = NULL;
static GtkWidget *preview_progressbar = NULL;
static gboolean preview_mode;
static gboolean is_rgb;
static GimpDrawable *drawable;
static gdouble *gradient = NULL;
static gint gradient_p_width = -1;
static gboolean in_param_reset = TRUE;
static gboolean data_calculated = FALSE;

static gint    ix1, iy1, ix2, iy2;     /* Selected image size. */
static gint    bpp, has_alpha, alpha;
static glong   progress;
static gdouble max_progress;  /* saves typecasting to double */
static gint    minz, maxz;
static gint   *particle_data = NULL;
static gint   *dh = NULL;


MAIN()

static void
query(void)
{
    static GimpParamDef args[] = {
        { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
        { GIMP_PDB_IMAGE, "image", "Input image (unused)" },
        { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
        { GIMP_PDB_INT32, "seed", "Random seed" },
        { GIMP_PDB_FLOAT, "density", "Particle density" },
        { GIMP_PDB_FLOAT, "size_mean", "Mean particle size" },
        { GIMP_PDB_FLOAT, "size_range", "Particle size range" },
        { GIMP_PDB_FLOAT, "vert_mean", "Mean particle flight verticality" },
        { GIMP_PDB_FLOAT, "vert_range", "Particle verticality range" },
        { GIMP_PDB_FLOAT, "dir_strength", "Anisotropy effect strength" },
        { GIMP_PDB_FLOAT, "dir_angle", "Anisotropy angle (degrees)" },
        { GIMP_PDB_INT32, "low_born", "Enable low-born particles" },
        { GIMP_PDB_INT32, "sharper", "Make the resulting grains less round" },
        { GIMP_PDB_INT32, "cm", "Coloring method: { GRAYSCALE (0), GRADIENT (1) }" },
        { GIMP_PDB_INT32, "grad_invert", "Invert gradient" },
        { GIMP_PDB_INT32, "equalize_map", "Equalize color map to height distribution" },
        { GIMP_PDB_STRING, "grad", "Gradient used for coloring (when cm is GRADIENT)" },
    };

    gimp_install_procedure
        (DATA_KEY_VALS,
         "Generates a texture simulating particle deposition",
         "This plugin generates a tileable texture by emulating deposition "
         "of particles on a surface.  The resulting height field is then "
         "visualized.",
         "Yeti",
         "Yeti",
         "Dec 2005",
         N_("<Image>/Filters/Render/Nature/_Particle Deposition..."),
         "RGB*, GRAY*",
         GIMP_PLUGIN,
         G_N_ELEMENTS(args), 0, args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam *param,
    gint *nreturn_vals,
    GimpParam **return_vals)
{
    static GimpParam values[1];
    GimpRunMode run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;

    run_mode = param[0].data.d_int32;

    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    textdomain(GETTEXT_PACKAGE);

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    drawable = gimp_drawable_get(param[2].data.d_drawable);
    is_rgb = gimp_drawable_is_rgb(drawable->drawable_id);
    has_alpha = gimp_drawable_has_alpha(drawable->drawable_id);

    pvals = pvals_defaults;
    switch (run_mode) {
        case GIMP_RUN_INTERACTIVE:
        gimp_get_data(DATA_KEY_VALS, &pvals);

        pvals_old = pvals;
        if (!plugin_dialog()) {
            gimp_drawable_detach(drawable);
            return;
        }
        break;

        case GIMP_RUN_NONINTERACTIVE:
        if (nparams != 17)
            status = GIMP_PDB_CALLING_ERROR;
        else {
            pvals.seed = (gint)param[3].data.d_int32;
            pvals.density = (gdouble)param[4].data.d_float;
            pvals.size_mean = (gdouble)param[5].data.d_float;
            pvals.size_range = (gdouble)param[6].data.d_float;
            pvals.vert_mean = (gdouble)param[7].data.d_float;
            pvals.vert_range = (gdouble)param[8].data.d_float;
            pvals.dir_strength = (gdouble)param[9].data.d_float;
            pvals.dir_angle = (gdouble)param[10].data.d_float;
            pvals.low_born = (gboolean)param[11].data.d_int32;
            pvals.sharper = (gboolean)param[12].data.d_int32;
            pvals.cm = (ColoringMethod)param[13].data.d_int32;
            pvals.grad_invert = (gboolean)param[14].data.d_int32;
            pvals.equalize_map = (gboolean)param[15].data.d_int32;
            g_strlcpy(pvals.grad, (gchar*)param[16].data.d_string,
                      sizeof(pvals.grad));
        }
        break;

        case GIMP_RUN_WITH_LAST_VALS:
        gimp_get_data(DATA_KEY_VALS, &pvals);
        break;

        default:
        break;
    }

    if (status == GIMP_PDB_SUCCESS) {
        if (is_rgb || gimp_drawable_is_gray(drawable->drawable_id)) {
            gimp_progress_init(_("Particle Deposition..."));

            preview_mode = FALSE;
            in_param_reset = FALSE;
            execute(drawable, RECOMPUTE_ALL);

            if (run_mode != GIMP_RUN_NONINTERACTIVE)
                gimp_displays_flush();

            if (run_mode == GIMP_RUN_INTERACTIVE
                || run_mode == GIMP_RUN_WITH_LAST_VALS) {
                g_strlcpy(pvals.grad, gimp_context_get_gradient(),
                          sizeof(pvals.grad));
                gimp_set_data(DATA_KEY_VALS, &pvals, sizeof(PluginValues));
            }
        }
        else
            status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;
    gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline void
yeti_table_attach_to_row(GtkWidget *table,
                         GtkWidget *child,
                         gint row)
{
    gtk_table_attach(GTK_TABLE(table), child, 0, GTK_TABLE(table)->ncols,
                     row, row+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline void
yeti_table_attach_single_cell(GtkWidget *table,
                              GtkWidget *child,
                              gint column,
                              gint row)
{
    gtk_table_attach(GTK_TABLE(table), child, column, column+1, row, row+1,
                     GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline GtkWidget*
yeti_check_button_new_with_label(const gchar *name,
                                 gint *value,
                                 GCallback cb,
                                 gpointer data)
{
    GtkWidget *check;

    check = gtk_check_button_new_with_mnemonic(name);
    g_signal_connect(check, "toggled",
                     G_CALLBACK(gimp_toggle_button_update), value);
    if (cb)
        g_signal_connect_swapped(check, "toggled", cb, data);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), *value);

    return check;
}

static inline GtkWidget*
yeti_preview_new_with_progress(GtkWidget *hbox,
                               GimpDrawable *drawable,
                               GCallback callback,
                               GtkWidget **progressbar,
                               GtkWidget **button)
{
    GtkWidget *frame, *vbox, *image;
    GdkPixbuf *pixbuf;

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,
                            FALSE,
                            BITS_PER_SAMPLE,
                            PREVIEW_SIZE,
                            PREVIEW_SIZE);
    gdk_pixbuf_fill(pixbuf, 0);

    image = gtk_image_new_from_pixbuf(pixbuf);
    g_object_unref(pixbuf);
    gtk_container_add(GTK_CONTAINER(frame), image);

    *progressbar = gtk_progress_bar_new();
    gtk_widget_set_size_request(*progressbar, PREVIEW_SIZE, 8);
    gtk_box_pack_start(GTK_BOX(vbox), *progressbar, FALSE, FALSE, 0);

    *button = gtk_button_new_with_mnemonic(_("_Update Preview"));
    gtk_box_pack_start(GTK_BOX(vbox), *button, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(*button), "clicked",
                     G_CALLBACK(callback), drawable);

    return image;
}

static inline GtkObject*
yeti_scale_entry_new_double(const gchar *name,
                            GtkWidget *table,
                            GtkSizeGroup *sgroup,
                            gint row,
                            double *value,
                            gdouble min,
                            gdouble max,
                            gdouble step,
                            gdouble page,
                            gint digits)
{
    GtkObject *adj;

    *value = CLAMP(*value, min, max);
    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, step, page, digits,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(gimp_double_adjustment_update), value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}

static inline void
yeti_coloring_method_attach(GtkWidget *table,
                            gint *row,
                            const gchar *gradlabel,
                            GtkWidget **grad,
                            GtkWidget **cm_gray,
                            GtkWidget **cm_grad,
                            ColoringMethod *cm)
{
    GtkRadioButton *group;
    GtkWidget *align;

    if (!is_rgb)
        *cm = COLORING_METHOD_GRAY;

    *grad = gimp_gradient_select_widget_new(gradlabel,
                                            gimp_context_get_gradient(),
                                            gradient_changed,
                                            NULL);
    align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
    gtk_container_add(GTK_CONTAINER(align), *grad);
    yeti_table_attach_single_cell(table, align, 1, *row+1);
    gtk_widget_set_sensitive(*grad, *cm == COLORING_METHOD_GRADIENT);

    *cm_gray = gtk_radio_button_new_with_mnemonic(NULL, _("Gra_yscale"));
    group = GTK_RADIO_BUTTON(*cm_gray);
    g_signal_connect(*cm_gray, "toggled",
                     G_CALLBACK(coloring_method_changed),
                     GUINT_TO_POINTER(COLORING_METHOD_GRAY));
    yeti_table_attach_single_cell(table, *cm_gray, 0, *row);

    *cm_grad = gtk_radio_button_new_with_mnemonic_from_widget(group,
                                                              _("_Gradient"));
    if (*cm == COLORING_METHOD_GRADIENT)
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*cm_grad), TRUE);
    g_signal_connect(*cm_grad, "toggled",
                     G_CALLBACK(coloring_method_changed),
                     GUINT_TO_POINTER(COLORING_METHOD_GRADIENT));
    yeti_table_attach_single_cell(table, *cm_grad, 0, *row+1);
    if (!is_rgb)
        gtk_widget_set_sensitive(*cm_grad, FALSE);

    *row += 2;
}
/* }}} */

static gboolean
plugin_dialog(void)
{
    GtkWidget *dialog, *main_vbox, *hbox, *table, *align;
    GtkSizeGroup *sgroup;
    gint row, run;
    gboolean ok;

    preview_mode = TRUE;
    in_param_reset = TRUE;
    gimp_ui_init(PROCEDURE_NAME, TRUE);

    dialog = gimp_dialog_new(_("Particle Deposition"), PROCEDURE_NAME,
                             NULL, 0,
                             gimp_standard_help_func,
                             "filters/particle_deposition.html",

                             _("Re_vert"), RESPONSE_REVERT,
                             GIMP_STOCK_RESET, RESPONSE_RESET,
                             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                             GTK_STOCK_OK, GTK_RESPONSE_OK,
                             NULL);

    main_vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 6);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), main_vbox,
                       TRUE, TRUE, 0);

    hbox = gtk_hbox_new(FALSE, 8);
    gtk_box_pack_start(GTK_BOX(main_vbox), hbox, TRUE, TRUE, 0);

    preview_image = yeti_preview_new_with_progress(hbox, drawable,
                                                   G_CALLBACK(execute_all),
                                                   &preview_progressbar,
                                                   &pctrl.update);
    preview = gtk_image_get_pixbuf(GTK_IMAGE(preview_image));

    table = gtk_table_new(3, 4, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    row = 0;

    align = gtk_alignment_new(0.0, 0.0, 1.0, 0.0);
    gtk_container_add(GTK_CONTAINER(align), table);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);

    yeti_coloring_method_attach(table, &row, _("Particle Deposition Gradient"),
                                &pctrl.grad, &pctrl.cm_gray, &pctrl.cm_grad,
                                &pvals.cm);

    pctrl.grad_invert
        = yeti_check_button_new_with_label(_("_Invert"),
                                           &pvals.grad_invert,
                                           G_CALLBACK(execute_color),
                                           drawable);
    yeti_table_attach_to_row(table, pctrl.grad_invert, row);
    row++;

    pctrl.equalize_map
        = yeti_check_button_new_with_label(_("_Equalize mapping"),
                                           &pvals.equalize_map,
                                           G_CALLBACK(execute_color),
                                           drawable);
    yeti_table_attach_to_row(table, pctrl.equalize_map, row);
    row++;

    table = gtk_table_new(10, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_box_pack_start(GTK_BOX(main_vbox), table, FALSE, FALSE, 0);
    row = 0;

    sgroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    pctrl.seed = gimp_random_seed_new(&pvals.seed, &pvals.random_seed);
    gimp_table_attach_aligned(GTK_TABLE(table), 0, row,
                              _("Random seed:"), 0.0, 0.5,
                              pctrl.seed, 2, TRUE);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.density
        = yeti_scale_entry_new_double(_("Particle _density"),
                                      table, sgroup, row, &pvals.density,
                                      1.0, 3000.0, 1.0, 100.0, 0);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.size_mean
        = yeti_scale_entry_new_double(_("Mean particle _size:"),
                                      table, sgroup, row, &pvals.size_mean,
                                      0.01, 1.0, 0.01, 0.1, 2);
    row++;

    pctrl.size_range
        = yeti_scale_entry_new_double(_("_Particle size range:"),
                                      table, sgroup, row, &pvals.size_range,
                                      0.01, 1.0, 0.01, 0.1, 2);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.vert_mean
        = yeti_scale_entry_new_double(_("_Mean flight verticallity:"),
                                      table, sgroup, row, &pvals.vert_mean,
                                      0.01, 1.0, 0.01, 0.1, 2);
    row++;

    pctrl.vert_range
        = yeti_scale_entry_new_double(_("_Flight verticallity range:"),
                                      table, sgroup, row, &pvals.vert_range,
                                      0.01, 2.0, 0.01, 0.1, 2);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.dir_strength
        = yeti_scale_entry_new_double(_("_Anisotropy strength"),
                                      table, sgroup, row, &pvals.dir_strength,
                                      0.0, 1.0, 0.01, 0.1, 2);
    row++;

    pctrl.dir_angle
        = yeti_scale_entry_new_double(_("Aniso_tropy angle"),
                                      table, sgroup, row, &pvals.dir_angle,
                                      -180.0, 180.0, 1.0, 15.0, 1);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.low_born = yeti_check_button_new_with_label(_("Enable _low-born "
                                                        "particles"),
                                                      &pvals.low_born,
                                                      NULL, NULL);
    yeti_table_attach_to_row(table, pctrl.low_born, row);
    row++;

    pctrl.sharper = yeti_check_button_new_with_label(_("S_harper shapes"),
                                                     &pvals.sharper,
                                                     NULL, NULL);
    yeti_table_attach_to_row(table, pctrl.sharper, row);
    row++;

    g_object_unref(sgroup);
    gtk_widget_show_all(dialog);
    in_param_reset = FALSE;

    ok = FALSE;
    do {
        run = gimp_dialog_run(GIMP_DIALOG(dialog));
        switch (run) {
            case RESPONSE_REVERT:
            refresh_controls(&pvals_old);
            break;

            case RESPONSE_RESET:
            refresh_controls(&pvals_defaults);
            break;

            default:
            ok = TRUE;
            break;
        }
    } while (!ok);

    gtk_widget_destroy(dialog);

    return run == GTK_RESPONSE_OK;
}

static void
coloring_method_changed(GtkWidget *widget,
                        gpointer data)
{
    ColoringMethod cm;

    if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
        return;

    cm = GPOINTER_TO_UINT(data);
    if (pvals.cm == cm)
        return;
    gtk_widget_set_sensitive(pctrl.grad, cm == COLORING_METHOD_GRADIENT);
    pvals.cm = cm;
    execute_color(drawable);
}

static void
gradient_changed(const gchar *name,
                 gint width,
                 const gdouble *grad_data,
                 gboolean dialog_closing,
                 gpointer user_data)
{
    gimp_context_set_gradient(name);
    gradient = g_realloc(gradient, width*sizeof(gdouble));
    memcpy(gradient, grad_data, width*sizeof(gdouble));
    gradient_p_width = width/4;

    execute_color(drawable);
}

/* though this is obscure, we use it to reset/revert all values _and_ refresh
   the dialog controls */
static void
refresh_controls(const PluginValues *new_pvals)
{
    in_param_reset = TRUE;

    pvals.density = new_pvals->density;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.density),
                             pvals.density);

    pvals.size_mean = new_pvals->size_mean;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.size_mean),
                             pvals.size_mean);
    pvals.size_range = new_pvals->size_range;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.size_range),
                             pvals.size_range);

    pvals.vert_mean = new_pvals->vert_mean;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.vert_mean),
                             pvals.vert_mean);
    pvals.vert_range = new_pvals->vert_range;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.vert_range),
                             pvals.vert_range);

    pvals.dir_strength = new_pvals->dir_strength;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.dir_strength),
                             pvals.dir_strength);
    pvals.dir_angle = new_pvals->dir_angle;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.dir_angle),
                             pvals.dir_angle);

    pvals.low_born = new_pvals->low_born;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.low_born),
                                 pvals.low_born);
    pvals.sharper = new_pvals->sharper;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.sharper),
                                 pvals.sharper);

    pvals.cm = new_pvals->cm;
    if (!is_rgb)
        pvals.cm = COLORING_METHOD_GRAY;
    switch (pvals.cm) {
        case COLORING_METHOD_GRAY:
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.cm_gray), TRUE);
        break;

        case COLORING_METHOD_GRADIENT:
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.cm_grad), TRUE);
        break;
    }
    gtk_widget_set_sensitive(pctrl.grad, pvals.cm == COLORING_METHOD_GRADIENT);

    pvals.seed = new_pvals->seed;
    gtk_adjustment_set_value(GIMP_RANDOM_SEED_SPINBUTTON_ADJ(pctrl.seed),
                             pvals.seed);

    pvals.grad_invert = new_pvals->grad_invert;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.grad_invert),
                                 pvals.grad_invert);

    pvals.equalize_map = new_pvals->equalize_map;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.equalize_map),
                                 pvals.equalize_map);

    /* unable to update: grad -- FIXME */
    in_param_reset = FALSE;
    execute(drawable, RECOMPUTE_COLOR);
}

/****************************************************************************/

#define TAUS_LCG(n) ((69069U * n) & 0xffffffffU)
#define TAUS1(x, s1, r1, m, s2) \
    ((((x << s1) & TAUS_MASK) ^ x) >> r1) ^ (((x & m) << s2) & TAUS_MASK)
#define TAUS_MASK 0xffffffffU
#define UNIT 0x400

static guint32 taus_state[4];

static inline guint32
taus_get(void)
{
    taus_state[0] = TAUS1(taus_state[0],  6U, 13U, 4294967294U, 18U);
    taus_state[1] = TAUS1(taus_state[1],  2U, 27U, 4294967288U,  2U);
    taus_state[2] = TAUS1(taus_state[2], 13U, 21U, 4294967280U,  7U);
    taus_state[3] = TAUS1(taus_state[3],  3U, 12U, 4294967168U, 13U);

    return taus_state[0] ^ taus_state[1] ^ taus_state[2] ^ taus_state[3];
}

static void
taus_seed(guint32 s)
{
    static const guint32 shift[] = { 2U, 8U, 16U, 128U };
    guint32 i;

    for (i = 0; i < 4; i++) {
        taus_state[i] = TAUS_LCG(i ? taus_state[i-1] : s);
        if (taus_state[i] < shift[i])
            taus_state[i] += shift[i];
    }
    for (i = 0; i < 10; i++)
        taus_get();
}

/****************************************************************************/

static void
execute_color(GimpDrawable *drawable)
{
    if (!data_calculated)
        return;

    execute(drawable, RECOMPUTE_COLOR);
}

static void
execute_all(GimpDrawable *drawable)
{
    gpointer *button, *toggle;

    toggle = GIMP_RANDOM_SEED_TOGGLE(pctrl.seed);
    /* This is dirty, so don't try anything when we fail to get the button */
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle))
        && (button = g_object_get_data(G_OBJECT(pctrl.seed), "button")))
        g_signal_emit_by_name(button, "clicked");

    execute(drawable, RECOMPUTE_ALL);
}

static void
plugin_execute(void)
{
    gint vz0, vzr, ps0, psr;
    gint i, j, hz, vx, vx0, vy, vy0, vz, rem, w, h, psize;
    guint64 niter, iter;
    gdouble phi;
    gint *z;

    taus_seed(pvals.seed);
    z = particle_data;
    w = ix2 - ix1;
    h = iy2 - iy1;
    niter = w*h*(guint64)pvals.density;
    max_progress = niter/PROGRESS_STEP;
    progress = 0;

    ps0 = MAX((gint)(UNIT*(pvals.size_mean - pvals.size_range)), 1);
    psr = (gint)(UNIT*(pvals.size_mean + pvals.size_range)) - ps0 + 1;

    vz0 = MAX((gint)(UNIT*(pvals.vert_mean - pvals.vert_range)), 1);
    vzr = (gint)(UNIT*(pvals.vert_mean + pvals.vert_range)) - vz0 + 1;

    memset(z, 0, w*h*sizeof(gint));
    maxz = 0;

    for (iter = 0; iter < niter; iter++) {
        i = taus_get() % h;  /* inital y-position */
        j = taus_get() % w;  /* inital x-position */
        /* genrate horizontal velocity direction */
        phi = 2.0*taus_get()/TAUS_MASK;
        phi = phi*(pvals.dir_strength*(phi - 1.0)*(phi - 2.0) + 1.0);
        if (pvals.sharper)
            phi = floor((phi + 0.2)/0.4)*2.0*G_PI/5.0 - G_PI/5.0;
        else
            phi *= G_PI;
        phi -= pvals.dir_angle*G_PI/180.0;

        psize = taus_get() % psr + ps0;  /* particle size */
        vx = vx0 = (gint)floor(UNIT*cos(phi) + 0.5);  /* x-velocity */
        vy = vy0 = (gint)floor(UNIT*sin(phi) + 0.5);  /* y-velocity */
        vz = taus_get() % vzr + vz0;  /* z-velocity */
        if (pvals.low_born)
            hz = MIN(z[i*w + j] + 120*vz, maxz + 3*vz/2);
        else
            hz = maxz + 3*vz/2;
        rem = 0;

        /* Flight iteration {{{ */
        if (vx >= 0) {
            if (vy >= 0) {
                if (vx >= vy) {
                    vz = vz*UNIT/vx;
                    while (hz > z[i*w + j] + vz/2) {
                        j++;
                        if (G_UNLIKELY(j == w))
                            j = 0;
                        rem += vy;
                        if (rem >= vx) {
                            i++;
                            if (G_UNLIKELY(i == h))
                                i = 0;
                            rem -= vx;
                        }
                        hz -= vz;
                    }
                }
                else {
                    vz = vz*UNIT/vy;
                    while (hz > z[i*w + j] + vz/2) {
                        i++;
                        if (G_UNLIKELY(i == h))
                            i = 0;
                        rem += vx;
                        if (rem >= vy) {
                            j++;
                            if (G_UNLIKELY(j == w))
                                j = 0;
                            rem -= vy;
                        }
                        hz -= vz;
                    }
                }
            }
            else {
                vy = -vy;
                if (vx >= vy) {
                    vz = vz*UNIT/vx;
                    while (hz > z[i*w + j] + vz/2) {
                        j++;
                        if (G_UNLIKELY(j == w))
                            j = 0;
                        rem += vy;
                        if (rem >= vx) {
                            i--;
                            if (G_UNLIKELY(i == -1))
                                i = h-1;
                            rem -= vx;
                        }
                        hz -= vz;
                    }
                }
                else {
                    vz = vz*UNIT/vy;
                    while (hz > z[i*w + j] + vz/2) {
                        i--;
                        if (G_UNLIKELY(i == -1))
                            i = h-1;
                        rem += vx;
                        if (rem >= vy) {
                            j++;
                            if (G_UNLIKELY(j == w))
                                j = 0;
                            rem -= vy;
                        }
                        hz -= vz;
                    }
                }
            }
        }
        else {
            vx = -vx;
            if (vy >= 0) {
                if (vx >= vy) {
                    vz = vz*UNIT/vx;
                    while (hz > z[i*w + j] + vz/2) {
                        j--;
                        if (G_UNLIKELY(j == -1))
                            j = w-1;
                        rem += vy;
                        if (rem >= vx) {
                            i++;
                            if (G_UNLIKELY(i == h))
                                i = 0;
                            rem -= vx;
                        }
                        hz -= vz;
                    }
                }
                else {
                    vz = vz*UNIT/vy;
                    while (hz > z[i*w + j] + vz/2) {
                        i++;
                        if (G_UNLIKELY(i == h))
                            i = 0;
                        rem += vx;
                        if (rem >= vy) {
                            j--;
                            if (G_UNLIKELY(j == -1))
                                j = w-1;
                            rem -= vy;
                        }
                        hz -= vz;
                    }
                }
            }
            else {
                vy = -vy;
                if (vx >= vy) {
                    vz = vz*UNIT/vx;
                    while (hz > z[i*w + j] + vz/2) {
                        j--;
                        if (G_UNLIKELY(j == -1))
                            j = w-1;
                        rem += vy;
                        if (rem >= vx) {
                            i--;
                            if (G_UNLIKELY(i == -1))
                                i = h-1;
                            rem -= vx;
                        }
                        hz -= vz;
                    }
                }
                else {
                    vz = vz*UNIT/vy;
                    while (hz > z[i*w + j] + vz/2) {
                        i--;
                        if (G_UNLIKELY(i == -1))
                            i = h-1;
                        rem += vx;
                        if (rem >= vy) {
                            j--;
                            if (G_UNLIKELY(j == -1))
                                j = w-1;
                            rem -= vy;
                        }
                        hz -= vz;
                    }
                }
            }
        }
        /* }}} */

        if (hz < z[i*w + j]) {
            /* Trace back one step, speed optimization is not as crucial as
             * above */
            if (vx > vy) {
                j = (j + (vx0 > 0 ? w-1 : 1)) % w;
                rem -= vy;
                if (rem < 0)
                    i = (i + (vy0 > 0 ? h-1 : 1)) % h;
            }
            else {
                i = (i + (vy0 > 0 ? h-1 : 1)) % h;
                rem -= vx;
                if (rem < 0)
                    j = (j + (vx0 > 0 ? w-1 : 1)) % w;
            }
        }
        z[i*w + j] += psize;
        if (z[i*w + j] > maxz)
            maxz = z[i*w + j];

        if (iter % PROGRESS_STEP == PROGRESS_STEP-1) {
            progress++;
            if (preview_mode) {
                gtk_progress_bar_set_fraction
                    (GTK_PROGRESS_BAR(preview_progressbar),
                     progress/max_progress);
                while (gtk_events_pending())
                    gtk_main_iteration_do(FALSE);
            }
            else
                gimp_progress_update(progress/max_progress);
        }
    }

    minz = maxz;
    if (dh)
        g_free(dh);
    dh = g_new0(gint, maxz+1);
    for (i = 0; i < w*h; i++) {
        dh[z[i]]++;
        if (z[i] < minz)
            minz = z[i];
    }
    for (i = 1; i <= maxz; i++)
        dh[i] += dh[i-1];

    if (preview_mode) {
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(preview_progressbar),
                                      0.0);
        while (gtk_events_pending())
            gtk_main_iteration_do(FALSE);
    }

    data_calculated = TRUE;
}

static void
execute(GimpDrawable *drawable,
        guint recompute)
{
    gint chmax;

    if (in_param_reset)
        return;

    if (preview_mode)
        gtk_widget_set_sensitive(pctrl.update, FALSE);

    /* compute sizes and other stuff of this kind */
    alpha = -1000; /* to cause a segfault ;-> */
    if (preview_mode) {
        ix1 = iy1 = 0;
        ix2 = gdk_pixbuf_get_width(preview);
        iy2 = gdk_pixbuf_get_height(preview);
        bpp = gdk_pixbuf_get_n_channels(preview);
        if (!particle_data)
            particle_data = g_new(gint, ix2*iy2);
    }
    else {
        gimp_drawable_mask_bounds(drawable->drawable_id,
                                  &ix1, &iy1, &ix2, &iy2);
        bpp = gimp_drawable_bpp(drawable->drawable_id);
        if (particle_data)
            g_free(particle_data);
        particle_data = g_new(gint, (ix2 - ix1)*(iy2 - iy1));
        if (has_alpha)
            alpha = bpp-1;
    }
    chmax = bpp;

    render_gradient();
    if ((recompute & RECOMPUTE_ALL))
        plugin_execute();
    commit(drawable, chmax);

    if (preview_mode) {
        gtk_widget_queue_draw(preview_image);
        gtk_widget_set_sensitive(pctrl.update, TRUE);
        /* http://bugzilla.gnome.org/show_bug.cgi?id=56070 workaround */
        gtk_widget_hide(pctrl.update);
        gtk_widget_show(pctrl.update);
    }
    else {
        gimp_drawable_flush(drawable);
        gimp_drawable_merge_shadow(drawable->drawable_id, TRUE);
        gimp_drawable_update(drawable->drawable_id,
                             ix1, iy1, (ix2 - ix1), (iy2 - iy1));
    }
}

/* give the abstract values some visual interpretation */
static void
commit(GimpDrawable *drawable,
       gint chmax)
{
    gpointer pr;
    GimpPixelRgn dest_rgn;

    if (preview_mode) {
        dest_rgn.x = dest_rgn.y = 0;
        dest_rgn.w = dest_rgn.h = PREVIEW_SIZE;
        dest_rgn.bpp = gdk_pixbuf_get_n_channels(preview);
        dest_rgn.rowstride = gdk_pixbuf_get_rowstride(preview);
        dest_rgn.data = gdk_pixbuf_get_pixels(preview);
        commit_one_tile(&dest_rgn, chmax, gradient_p_width);
    }
    else {
        gimp_pixel_rgn_init(&dest_rgn, drawable,
                            ix1, iy1, (ix2 - ix1), (iy2 - iy1),
                            TRUE, TRUE);
        /* iterate by tiles */
        for (pr = gimp_pixel_rgns_register(1, &dest_rgn);
             pr != NULL;
             pr = gimp_pixel_rgns_process(pr)) {
            commit_one_tile(&dest_rgn, chmax, GRADIENT_SAMPLE_SIZE);
        }
    }
}

/* give the abstract values some visual interpretation (one tile) */
static void
commit_one_tile(GimpPixelRgn *rgn,
                gint chmax,
                gint gradwidth)
{
    static const gint check_colors[] = {
        CHANNEL_MAX_VALUE*GIMP_CHECK_DARK,
        CHANNEL_MAX_VALUE*GIMP_CHECK_LIGHT,
    };

    gint index = 0, channel, chc, v;
    guint x, y;
    guchar *p;
    gdouble r = 0, f;

    for (y = rgn->y; y < rgn->y + rgn->h; y++) {
        for (x = rgn->x; x < rgn->x + rgn->w; x++) {
            /* draw the pixel */
            p = rgn->data + (y - rgn->y)*rgn->rowstride + (x - rgn->x)*rgn->bpp;
            v = particle_data[(y - iy1)*(ix2 - ix1) + x - ix1];
            if (pvals.equalize_map)
                r = (gdouble)(dh[v] - dh[0])/(dh[maxz] - dh[0]);
            else
                r = (gdouble)(v - minz)/(maxz - minz);
            if (pvals.grad_invert)
                r = 1.0 - r;
            if (pvals.cm == COLORING_METHOD_GRADIENT)
                index = r*(gradwidth - 1);
            for (channel = 0; channel < chmax; channel++) {
                if (pvals.cm == COLORING_METHOD_GRADIENT) {
                    p[channel] = CHANNEL_MAX_VALUE*gradient[4*index + channel];
                    /* add checker in preview mode */
                    if (preview_mode && has_alpha) {
                        f = gradient[4*index + 3];
                        chc = (x - ix1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE
                            + (y - iy1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE;
                        p[channel] = p[channel]*f
                                     + (1.0 - f)*check_colors[chc % 2];
                    }
                }
                else {
                    if (channel == alpha)
                        p[channel] = CHANNEL_MAX_VALUE;
                    else
                        p[channel] = CHANNEL_MAX_VALUE*r;
                }
            }
        }
    }
}

static void
render_gradient(void)
{
    gint dummy;

    if (pvals.cm != COLORING_METHOD_GRADIENT)
        return;

    if (preview_mode) {
        /* don't regenerate what we already have */
        if (gradient == NULL) {
            gradient_p_width = PREVIEW_SIZE;
            gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                              gradient_p_width,
                                              FALSE,
                                              &dummy, &gradient);
        }
    }
    else {
        /* final rendering, done only once at the end */
        g_free(gradient);
        gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                          GRADIENT_SAMPLE_SIZE,
                                          FALSE,
                                          &dummy, &gradient);
    }
}

